VERSION = "0.1.1"
LICENSE = "Apache 2.0"

SLEEP_MIN = 0.2  # Enforce minimum wait time between API calls (seconds)
PUBLIC_SERVER_BASE = "https://optipyzer-api.herokuapp.com"
LOCAL_SERVER_BASE = "http://localhost:8000"

SESSION_HDRS = {
	'application': 'Optipyzer',
	'User-Agent': 'https://github.com/NLeRoy917/optipyzer.com',
	'Content-Type': 'application/json'
}

VALID_AMINO_ACIDS = "ARNDCQEGHILKMFPSTWYV"
VALID_SEQ_TYPES = [
    'dna',
    'protein'
]


import re
from const import VALID_AMINO_ACIDS

def clean_seq(s: str) -> str:
    # remove all newlines and whitespace
    return re.sub(r"\s+", "", s)

def verify_dna(seq: str):
    # clean seqeunce
    seq = clean_seq(seq)

    # verify that the sequeryuence length
    # is divisible by 3 (codons)
    if len(seq) % 3 != 0:
        raise ValueError("DNA sequence must be devisible by 3")
    
    # verify all characters are
    # ATGC
    for base, i in zip(seq, range(len(seq))):
        if base.upper() not in "ATGC":
            raise ValueError(f"Invalid base '{base}' in query at position: {i}")
    
    return None

def verify_protein(seq: str):
    # clean seqeunce
    seq = clean_seq(seq)

    # verify each residue is valid
    for aa, i in zip(seq, range(len(seq))):
        if aa.upper() not in VALID_AMINO_ACIDS:
            raise ValueError(f"Invalid residue '{aa}' in query at position: {i}")
    
    return None


from typing import Dict
import requests
import time
from const import LOCAL_SERVER_BASE, PUBLIC_SERVER_BASE, SESSION_HDRS, SLEEP_MIN
from log import _LOGGER
from helpers import verify_dna, verify_protein

# return types
from requests import Response
from models import SearchResult
from const import VALID_SEQ_TYPES
from models import CodonUsage, OptimizationResult


class api():
	"""Python interface for the Optipyzer web API."""

	_session = requests.Session()
	_session.headers = SESSION_HDRS

	def __init__(self, local: bool = False, timeout: int = 1000, sleep_time: float = 0.5):
		"""Initialize an optipyzer interface"""

		# determine environment
		if local: self.api_base = LOCAL_SERVER_BASE
		else: self.api_base = PUBLIC_SERVER_BASE

		# set params
		self.timeout = timeout
		if sleep_time < SLEEP_MIN:
			_LOGGER.warn(
				f"Minimum sleep time of {SLEEP_MIN} sec required \
				({sleep_time} sec was supplied). Setting to {SLEEP_MIN} sec"
			)
			self.sleep_time = SLEEP_MIN
		else:
			self.sleep_time = sleep_time

	def _make_request(self, path: str, method: str = 'GET', params_: dict = {}, body_: dict = {}) -> Response:
		"""Make a request to the API"""
		
		# generate the URI
		uri = self.api_base + path

		try:
			response = self._session.request(
				method,
				uri,
				params=params_,
				json=body_
			)
		except requests.Timeout:
			_LOGGER.warn("Request timeout raised.")

		except requests.RequestException as e:
			_LOGGER.error(f"Exception raised during request to {uri}: {e}")

		# check response status
		if response.status_code != 200:
			_LOGGER.error(f"Request failed with status code: {response.status_code}")
			_LOGGER.error(f"{response.json()}")

		# Enforce rate limiting
		time.sleep(max(SLEEP_MIN, self.sleep_time))

		return response

	def search(self, name: str, limit: int = 50) -> SearchResult:
		"""Search for an organism given it's name"""
		result = self._make_request(
			"/species/search",
			params_={
				"name": name,
				"limit": limit
			}
		)
		search_results = result.json()
		return search_results

	def optimize(self, seq: str, weights: Dict[str, int], seq_type: str = 'dna') -> OptimizationResult:
		"""Optimize a sequence given specific organism weights"""
		# force seq_type lower
		seq_type = seq_type.lower()
		if seq_type not in VALID_SEQ_TYPES:
			raise ValueError(f"Invalid sequence type: {seq_type}")
		
		# validate that the sequences are valid
		if seq_type == 'dna': verify_dna(seq)
		else: verify_protein(seq)

		# make optimization request
		result = self._make_request(
			f"/optimize/{seq_type}",
			method="POST",
			body_={
				'seq': seq,
				'weights': weights
			}
		)
		return result.json()

	def pull_codons(self, org_id: str) -> CodonUsage:
		"""Pull codon usage data for a specific organism"""
		result = self._make_request(
			f"/species/{org_id}/codons",
		)
		return result.json()
		


if __name__ == '__main__':

	# test code goes here
	op = api(local=True)

	# search for e coli
	result = op.search(name='Escherichia Coli')
	orgs = result['organisms']
	print(orgs)

	# optimize a sequence
	seq = "ATGCGTACTAGTCAGTCAGACTGACTG"
	weights = {
	    "16815": 1,
	    "122563": 2
	}
	result = op.optimize(
		seq, weights, seq_type='dna'
	)

	# pull codons for e coli
	codon_usage = op.pull_codons("16815")
	print(codon_usage)












import logging

# init a logger
_LOGGER = logging.getLogger("optipyzer")
_LOGGER.setLevel(logging.DEBUG)

# generate a console-level logger
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)

# set formatting
formatter = logging.Formatter('%(levelname)s - %(message)s')
ch.setFormatter(formatter)

# add the handler
_LOGGER.addHandler(ch)


# Optipyzer API
# Nathan J. LeRoy
# See LICENSE for details

import sys

from .const import LICENSE, VERSION
assert sys.version_info[0] == 3, "Optipyzer requires Python 3.0 or greater"

"""A library that provides a Python interface to the Optipyzer API"""

__author__ = 'Nathan J. LeRoy'
__url__ = 'https://github.com/NLeRoy917/optipyzer.com'
__description__ = 'A Python wrapper around the Optipyzer API'
__license__ = LICENSE
__version__ = VERSION


from typing import List, Dict
from pydantic import BaseModel

class Organism(BaseModel):
    org_id: int
    division: str
    assembly: str
    taxid: int
    species: str
    organelle: str
    translation_table: int
    num_CDS: int
    num_codons: int
    GC_perc: float
    GC1_perc: float
    GC2_perc: float
    GC3_perc: float

class OptimizationResult(BaseModel):
    query: str
    weights: Dict[str, int]
    seq_type: str
    peptide_seq: str
    dna_seq: str
    stop_codon: int
    optimized_sd: str
    min_difference_sumsquares: float
    best_expression_sd: Dict[str, float]
    optimized_ad: str
    min_difference_absvalue: float
    best_expression_ad: Dict[str, float]

class SearchResult(BaseModel):
    num_results: int
    organisms: List[Organism]
    search_query: str

class CodonUsage(BaseModel):
    organism: Organism
    counts: Dict[str, Dict[str, int]]
    codon_usage: Dict[str, Dict[str, float]]



import setuptools
from optipyzer.const import VERSION

# create long desc.
with open("README.md", "r") as fh:
    long_description = fh.read()

# get requirements
requirements = []
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name='optipyzer',
    version=VERSION,
    description='Codon optimize DNA and Protein sequences for multiple species.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license="Apache 2.0",
    author='Nathan LeRoy',
    author_email='NLeRoy917@gmail.com',
    url='https://github.com/NLeRoy917/optipyzer',
    keywords='codon optimization dna protein biology bioinformatics',
    packages=setuptools.find_packages(),
    install_requires=requirements,
    classifiers=[
        'Topic :: Software Development :: Libraries',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ]
)